/*
 * Copyright (C) 2006 Alexander Papst
 *
 * Original concept/Algorithm - By Brian (http://scriptasylum.com/bgaudiodr/)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "pz.h"

#define HEADER ttk_screen->wy
#define HWH (ttk_screen->h-ttk_screen->wy)

static PzModule *module;
static PzConfig *tictactux_config;
static PzWindow *window;
static ttk_color grid_col, ex_col, oh_col, sel_col_p1, sel_col_p2;
static int ipod_human; 
static const char *tictactux_options[] = {"iPod", "Human", 0};

static int w, h;
static int xw1, xw2, yh1, yh2, ydist, fhw, c, s;
static unsigned int cur_button = 1;

static unsigned int level = 1;
static int myway, mynextmove;
static int tmp, done, iswon;
/*
static int ipodwins = [0, 0, 0, 0];
static int playerwins = [0, 0, 0, 0];
static int draws = [0, 0, 0, 0];
 */
static int playerstarts = 1;
static int whichplayer = 0;

// 11 | 12 | 13
//--------------
// 21 | 22 | 23
//--------------
// 31 | 32 | 33

static int moves[34];
static int game[9];

static int choices[9] = {11, 12, 13, 21, 22, 23, 31, 32, 33}; 
static int corners[4] = {11, 13, 31, 33}; 

static int ways[8][4] = {
	{0, 11, 12, 13}, // 0
	{0, 21, 22, 23}, // 1
	{0, 31, 32, 33}, // 2
	{0, 11, 21, 31}, // 3
	{0, 12, 22, 32}, // 4
	{0, 13, 23, 33}, // 5
	{0, 11, 22, 33}, // 6
	{0, 13, 22, 31}, // 7
};

//// Funtions:

//Drawing Stuff
static void output_header(int what);
static void draw_xo (ttk_surface srf);
static void draw_selection (ttk_surface srf);
static void show_level (ttk_surface srf);
static void draw_tictactux (PzWidget *wid, ttk_surface srf);
//Some kind of A.I.
static void ipodstrategy(int istowin);
static void selecCorner(int which);
static void ipoddontlose();
static void ipodrandom();
//rest
static void init ();
static int convert(int cur_field);
static void choosebutton(int cellnum, int direction);
static void setbutton(int cellnum);
static void ipodturn();
static void findwinner(int isplayer);

static void output_header(int what)
{
	if (what == 1) ttk_window_title(window, "Draw! :-|");
	else if (what == 2) ttk_window_title(window, "iPod wins! :-(");
	else if (what == 3) ttk_window_title(window, "Player wins! :-)"); 
	else if (what == 4) ttk_window_title(window, "Player 1 wins! :-)"); 
	else if (what == 5) ttk_window_title(window, "Player 2 wins! :-)"); 
}

static void draw_xo (ttk_surface srf)
{
	static int i;
	static unsigned int col, row;
	for (i=11; i<=33; i++) {
		
		if ((i == 11) || (i == 21) || (i == 31)) row = 0;
		if ((i == 12) || (i == 22) || (i == 32)) row = 1;
		if ((i == 13) || (i == 23) || (i == 33)) row = 2;
		
		if ((i == 11) || (i == 12) || (i == 13)) col = 0;
		if ((i == 21) || (i == 22) || (i == 23)) col = 1;
		if ((i == 31) || (i == 32) || (i == 33)) col = 2;

		if (moves[i] == 1) { // usr X
			ttk_line (srf, xw1-1+s+((ydist/3)*row), yh1+s+((ydist/3)*col), xw1-1+s+s+((ydist/3)*row), yh1+s+s+((ydist/3)*col), ex_col);
			ttk_line (srf, xw1-1+s+((ydist/3)*row)+1, yh1+s+((ydist/3)*col), xw1-1+s+s+((ydist/3)*row)+1, yh1+s+s+((ydist/3)*col), ex_col);
			ttk_line (srf, xw1-1+s+((ydist/3)*row)+2, yh1+s+((ydist/3)*col), xw1-1+s+s+((ydist/3)*row)+2, yh1+s+s+((ydist/3)*col), ex_col);
			
			ttk_line (srf, xw1-1+s+s+((ydist/3)*row), yh1+s+((ydist/3)*col), xw1-1+s+((ydist/3)*row), yh1+s+s+((ydist/3)*col), ex_col);
			ttk_line (srf, xw1-1+s+s+((ydist/3)*row)+1, yh1+s+((ydist/3)*col), xw1-1+s+((ydist/3)*row)+1, yh1+s+s+((ydist/3)*col), ex_col);
			ttk_line (srf, xw1-1+s+s+((ydist/3)*row)+2, yh1+s+((ydist/3)*col), xw1-1+s+((ydist/3)*row)+2, yh1+s+s+((ydist/3)*col), ex_col);
		}
		if (moves[i] == 2) { // iPod O
			ttk_ellipse (srf, xw1+c+((ydist/3)*row), yh1+c+((ydist/3)*col), s/2 , s/2, oh_col);
			ttk_ellipse (srf, xw1+c+((ydist/3)*row), yh1+c+((ydist/3)*col), s/2+1 , s/2+1, oh_col);
			if (w>=220) {
				ttk_ellipse (srf, xw1+c+((ydist/3)*row), yh1+c+((ydist/3)*col), s/2+2 , s/2+2, oh_col);
			}
		}
	}
}

static void draw_selection (ttk_surface srf)
{
	static ttk_color sel_col;
	if (whichplayer && ipod_human) sel_col = sel_col_p2; else sel_col = sel_col_p1;
	if (!iswon) {
		static unsigned int col, row;
		row = ((cur_button-1) % 3);
		if ((cur_button == 1) || (cur_button == 2) || (cur_button == 3)) col = 0;
		if ((cur_button == 4) || (cur_button == 5) || (cur_button == 6)) col = 1;
		if ((cur_button == 7) || (cur_button == 8) || (cur_button == 9)) col = 2;
		
		ttk_rect (srf, xw1+3+((ydist/3)*row), yh1+3+((ydist/3)*col), xw1+(ydist/3)-2+((ydist/3)*row), yh1+(ydist/3)-2+((ydist/3)*col), sel_col);
		ttk_rect (srf, xw1+4+((ydist/3)*row), yh1+4+((ydist/3)*col), xw1+(ydist/3)-3+((ydist/3)*row), yh1+(ydist/3)-3+((ydist/3)*col), sel_col);
	}
}

static void show_level (ttk_surface srf)
{
	if (ipod_human) {
		if (!iswon){
			if (whichplayer) ttk_text(srf, ttk_textfont, 2, h-2*ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "Player 2");
			else ttk_text(srf, ttk_textfont, 2, h-2*ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "Player 1");
		}
	} else {
		switch(level) { 
			case 0:
				ttk_text(srf, ttk_textfont, 2, h-2*ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "Level: 1");
				ttk_text(srf, ttk_textfont, 2, h-ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "The iPod is totally dumb.");
				break;
			case 1:
				ttk_text(srf, ttk_textfont, 2, h-2*ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "Level: 2");
				ttk_text(srf, ttk_textfont, 2, h-ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "The iPod is smarter.");
				break;
			case 2:
				ttk_text(srf, ttk_textfont, 2, h-2*ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "Level: 3");
				if (w == 138) {
					ttk_text(srf, ttk_textfont, 2, h-ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "You MAY win.");
				} else {
					ttk_text(srf, ttk_textfont, 2, h-ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "You MAY win (only when you start).");
				}
				break;
			case 3:
				ttk_text(srf, ttk_textfont, 2, h-2*ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "Level: 4");
				ttk_text(srf, ttk_textfont, 2, h-ttk_text_height(ttk_textfont)-HEADER, ttk_makecol(BLACK), "The iPod NEVER lose.");
				break;
			default: break;
		}
	}
}

static void draw_tictactux (PzWidget *wid, ttk_surface srf)
{	
	// Spielfeld zeichnen
	// Vertikal
	ttk_line (srf, xw1+(ydist/3), yh1, xw1+(ydist/3), yh2, grid_col);
	ttk_line (srf, xw1+(ydist*2/3), yh1, xw1+(ydist*2/3), yh2, grid_col);
	// Horizontal
	ttk_line (srf, xw1, yh1+(ydist/3), xw2, yh1+(ydist/3), grid_col);
	ttk_line (srf, xw1, yh1+(ydist*2/3), xw2, yh1+(ydist*2/3), grid_col);
	
	// Untere Linie
	ttk_fillrect (srf, 0, yh2+0.1*yh2, w, h, ttk_makecol(WHITE));
	ttk_line (srf, 0, yh2+0.1*yh2, w, yh2+0.1*yh2, ttk_makecol(BLACK));
	draw_xo(srf);
	draw_selection(srf);
	show_level(srf);
}

static void init ()
{
	static int i;
	iswon = 0;
	done = 0;
	cur_button = 1;
	
	moves[11] = 0;
	moves[12] = 0;
	moves[13] = 0;
	moves[21] = 0;
	moves[22] = 0;
	moves[23] = 0;
	moves[31] = 0;
	moves[32] = 0;
	moves[33] = 0;

	for (i=0; i<=8; i++) {
		game[i] = 0;
	}
	if (!playerstarts) {
		if (!ipod_human) ipodturn();
	}
}

static int convert(int cur_field)
{
	if (cur_field <= 3) cur_field += 10;
	else switch(cur_field) {
		case 4: cur_field = 21; break;
		case 5: cur_field = 22; break;
		case 6: cur_field = 23; break;
		case 7: cur_field = 31; break;
		case 8: cur_field = 32; break;
		case 9: cur_field = 33; break;
	default: break;
	}
	return cur_field;
}

static void choosebutton(int cellnum, int direction) 
{
	if (!iswon){
		if (direction == 1) {
			while(!moves[convert(cur_button)]==0) {
				cur_button++;
				if (cur_button <= 0) cur_button = 9;
				if (cur_button >= 10) cur_button = 1;
			}
		} else if (direction == (-1)) {
			while(!moves[convert(cur_button)]==0) {
				cur_button--;
				if (cur_button <= 0) cur_button = 9;
				if (cur_button >= 10) cur_button = 1;
			}
		}
	}
}

static void setbutton(int cellnum){
	if (!iswon){
		cellnum = convert(cellnum);
		if(moves[cellnum]==0){
			if (whichplayer && ipod_human) 	moves[cellnum]=2; else moves[cellnum]=1;
			game[done]=cellnum;
			done++;
			if (whichplayer && ipod_human) findwinner(0); else {
				findwinner(1);
			}
			whichplayer = !whichplayer;
		} 
	}	
}

static void ipodstrategy(int istowin) {
	static int str;
	static int n;
	if (istowin) str = 2; else str = 1;
	if (level>0) {
		for (n=0; n<=7; n++) {
			
			if ((moves[ways[n][1]] == str) && (moves[ways[n][2]] == str) && (moves[ways[n][3]] == 0)) {
				tmp = ways[n][3];
			}
			if ((moves[ways[n][1]] == str) && (moves[ways[n][3]] == str) && (moves[ways[n][2]] == 0)) {
				tmp = ways[n][2];
			}
			if ((moves[ways[n][2]] == str) && (moves[ways[n][3]] == str) && (moves[ways[n][1]] == 0)) {
				tmp = ways[n][1];
			}
		}
	}
}

static void selecCorner(int which) {
	if (which == 1) {
		do {
			tmp = corners[(int) (rand() % 4)];
		} while (moves[tmp] != 0);
	} else {
		tmp = corners[(int) (rand() % 4)];
	}
}

static void ipoddontlose() {
	static int i, dlta, op0, op1, op2;
	if (!playerstarts) {
		if (done == 0) {
			tmp = choices[2*(int) (rand() % 5)];
			if (tmp == 22) {
				myway = 1;
			} else {
				myway = 2;
			}
		} else if (done == 2) {
			if (myway == 1) {
				if (game[1] == 11 || game[1] == 13 || game[1] == 31 || game[1] == 33) {
					tmp = 44-game[1];
				} else {
					dlta = 22-game[1];
					op0 = 22+dlta+(10/dlta);
					op1 = 22+dlta-(10/dlta);
					switch((int) (rand() % 2)) {
						case 0: tmp = op0; break;
						case 1: tmp = op1; break;
						default: break;
					}
				}
			} else if (myway == 2) {
				if (game[1] == 22) {
					tmp = 44-game[0];
					myway = 21;
				} else if (game[1] == 11 || game[1] == 13 || game[1] == 31 || game[1] == 33) {
					selecCorner(1);
					myway = 22;
				} else {
					tmp = 22;
					myway = 23;
				}
			}
		} else if (done == 4) {
			if (myway == 22) {
				for (i=0; i<4; i++) {
					if (moves[corners[i]] == 0) {
						tmp = corners[i];
					}
				}
			} else if (myway == 23) {
				dlta = game[1]-game[0];
				op0 = 44-(game[1]+dlta);
				op1 = (op0+game[0])/2;
				switch((int) (rand() % 2)) {
					case 0: tmp = op0; break;
					case 1: tmp = op1; break;
					default: break;
				}
			} else if (myway == 1) {
				tmp = 44+game[2]-(2*game[3]);
			}
		}
	} else if (level == 3) {
		if (done == 1) {
			if (game[0] == 11 || game[0] == 13 || game[0] == 31 || game[0] == 33) {
				tmp = 22;
				myway = 1;
			} else if (game[0] == 22) {
				selecCorner(0);
				myway = 2;
			} else {
				tmp = 22;
				myway = 3;
			}
		} else if (done == 3) {
			if (myway == 1) {
				if (game[2] == 44-game[0]) {
					tmp = choices[1+(2*(int) (rand() % 4))];
				} else {
					tmp = 44-game[0];
				}
			} else if (myway == 2) {
				if (game[2] == 44-game[1]) {
					selecCorner(1);
				}
			} else if (myway == 3) {
				if (game[2] == 11 || game[2] == 13 || game[2] == 31 || game[2] == 33) {
					tmp = 44-game[2];
				}
				if (game[2] == 44-game[0]) {
					dlta = 22-game[2];
					tmp = 22+(10/dlta);
					mynextmove = tmp+dlta;
				} else {
					dlta = 22-game[0];
					op0 = game[0]+(10/dlta);
					op1 = game[0]-(10/dlta);
					op2 = game[2]+dlta;
					switch((int) (rand() % 3)) {
						case 0: tmp = op0; break;
						case 1: tmp = op1; break;
						case 2: tmp = op2; break;
						default: break;
					}
				}
			}
		} else if (done == 5 && myway == 3) {
			tmp = mynextmove;
		}
	}
}

static void findwinner(int isplayer) {
	static int me;
	static int n;
	if (isplayer) me = 1; else me = 2;
	for (n=0; n<=7; n++) {
		if ((moves[ways[n][1]] == me) && (moves[ways[n][2]] == me) && (moves[ways[n][3]] == me)) {
			iswon = 1;
			break;
		}
	}
	if (iswon) {
		if (isplayer) {
			//playerwins[level]++;
			playerstarts = 1;
			if (!ipod_human) output_header(3); else output_header(4);
		} else {
			//ipodwins[level]++;
			playerstarts = 0;
			if (!ipod_human) output_header(2); else output_header(5);
		}
	} else {
		if (done>8) {
			//draws[level]++;
			playerstarts = !playerstarts;
			iswon = 1;
			output_header(1);
		} else if (isplayer) {
			if (!ipod_human) ipodturn();
		}
	}
}

static void ipodrandom() {
	do {
		tmp = choices[(int) (rand() % 9)];
	} while (moves[tmp] != 0);
}

static void ipodturn() {
	tmp = 0;
	ipodstrategy(1);
	if (tmp == 0) {
		ipodstrategy(0);
	}
	if (tmp == 0 && level>1) {
		ipoddontlose();
	}
	if (tmp == 0) {
		ipodrandom();
	}
	if (tmp == 11) {
		cur_button++;
		choosebutton(cur_button,1);
	}
	moves[tmp] = 2;
	game[done] = tmp;
	
	done++;
	findwinner(0);
}

static int event_tictactux (PzEvent *ev) 
{
	int ret = TTK_EV_CLICK;

	switch(ev->type) {
	case PZ_EVENT_BUTTON_DOWN:
		switch (ev->arg) {
		case PZ_BUTTON_ACTION:
			if (iswon) {
				init();
				ttk_window_title(window, "Tic-Tac-Tux");
			} else setbutton(cur_button);
			ev->wid->dirty = 1;
			break;

		case PZ_BUTTON_NEXT:
			if (!ipod_human) {
				level = (level + 1) % 4; 
				if (iswon) {
					ttk_window_title(window, "Tic-Tac-Tux");
				}
				init();
				ev->wid->dirty = 1;
			}
			break;

		case PZ_BUTTON_PREVIOUS:
			if (!ipod_human) {
				level = (level - 1) % 4; 
				if (iswon) {
					ttk_window_title(window, "Tic-Tac-Tux");
				}
				init();
				ev->wid->dirty = 1;
			}
			break;

		case PZ_BUTTON_MENU:
			pz_close_window(ev->wid->win);
			break;

		default:
			ret = TTK_EV_UNUSED;
		}
		break;

	case PZ_EVENT_SCROLL:
		TTK_SCROLLMOD(ev->arg, 5);
		if (iswon) {
			init();
			ttk_window_title(window, "Tic-Tac-Tux");
		}
		cur_button += ev->arg;
		if (cur_button <= 0) cur_button = 9;
		if (cur_button >= 10) cur_button = 1;
		choosebutton(cur_button,ev->arg);
		ev->wid->dirty = 1;
		break;

	default:
		ret = TTK_EV_UNUSED;
		break;
	}
	return ret;
}

static PzWindow *new_tictactux_window()
{
	ipod_human = pz_get_int_setting(tictactux_config, 1);
	
	////Drawing init
	//resolution
	w = ttk_screen->w; 
	h = ttk_screen->h;
	//Color definitions
	if (ttk_ap_get("ttt.grid") == NULL) grid_col=ttk_makecol(0, 0, 0); else grid_col = ttk_ap_getx("ttt.grid")->color;
	if (ttk_ap_get("ttt.ex") == NULL) 
		if(ttk_screen->bpp == 16) ex_col=ttk_makecol(255, 0, 0); else ex_col=ttk_makecol(0, 0, 0);
	else ex_col = ttk_ap_getx("ttt.ex")->color;
	if (ttk_ap_get("ttt.oh") == NULL) 
		if(ttk_screen->bpp == 16) oh_col=ttk_makecol(0, 0, 255); else oh_col=ttk_makecol(0, 0, 0);
	else oh_col = ttk_ap_getx("ttt.oh")->color;
	if (ttk_ap_get("ttt.sel.ply1") == NULL) 
		if(ttk_screen->bpp == 16) sel_col_p1=ttk_makecol(255, 192, 0); else sel_col_p1=ttk_makecol(80, 80, 80);
	else sel_col_p1 = ttk_ap_getx("ttt.sel.ply1")->color;
	if (ttk_ap_get("ttt.sel.ply2") == NULL) 
		if(ttk_screen->bpp == 16) sel_col_p2=ttk_makecol(0, 255, 0); else sel_col_p2=ttk_makecol(192, 192, 192);
	else sel_col_p2 = ttk_ap_getx("ttt.sel.ply2")->color;
	//Grid
	yh1 = (HWH*5/100);
	yh2 = h-((h*1/10)+2*HEADER);
	ydist = yh2 -yh1;
	xw1 = w/2 - ydist/2;
	xw2 = xw1 + ydist;
	//Selection Frame
	fhw = (ydist/3)*0.9;
	//Size X, O
	s = (ydist/9);
	//Center X, O
	c = (ydist/6);
	
	////Statistics
	//ipodwins = [0, 0, 0, 0];
	//playerwins = [0, 0, 0, 0];
	//draws = [0, 0, 0, 0];
	
	srand((unsigned)time(NULL));
	playerstarts = 1;
	init();	

	window = pz_new_window ("Tic-Tac-Tux", PZ_WINDOW_NORMAL);
	pz_add_widget (window, draw_tictactux, event_tictactux);
	return pz_finish_window (window);
}

static void cleanup_tictactux() 
{
	pz_free_config(tictactux_config);
}

static int save_tictactux_config(TWidget *this, int key, int time)
{
	if(key == TTK_BUTTON_MENU){
		pz_save_config(tictactux_config); 
	}
	return ttk_menu_button(this, key, time);
}

static void init_tictactux() 
{
	module = pz_register_module ("tictactux", cleanup_tictactux);
	tictactux_config = pz_load_config(pz_module_get_cfgpath(module, ".conf"));
	pz_menu_add_action ("/Extras/Games/Tic-Tac-Tux/Start", new_tictactux_window);
	pz_menu_add_setting("/Extras/Games/Tic-Tac-Tux/Play against", 1, tictactux_config, tictactux_options);
	((TWidget *)pz_get_menu_item("/Extras/Games/Tic-Tac-Tux")->data)->button = save_tictactux_config;
}

PZ_MOD_INIT (init_tictactux)
